//@ts-check
"use strict";

/**
 * @typedef NPCName
 * @type {keyof typeof npcDescs}
 */

/**
 * @typedef NPCDesc
 * @property {NPCName} id
 * @property {boolean} [uncounted]
 */

/**
@type {{
  grandma: NPCDesc,
  farmhand: NPCDesc,
  jack: NPCDesc,
  blacksmith: NPCDesc,
  knightR: NPCDesc,
  knightG: NPCDesc,
  knightY: NPCDesc,
  knightB: NPCDesc,
  mom: NPCDesc,
  bro: NPCDesc,
  hermit: NPCDesc,
  sheep: NPCDesc,
  wizard: NPCDesc,
  bard: NPCDesc,
  conan: NPCDesc,
  broody: NPCDesc,
  gudul: NPCDesc,
  pup: NPCDesc,
  samurai: NPCDesc,
  captain: NPCDesc,
  onion: NPCDesc,
  leek: NPCDesc,
  salad: NPCDesc,
}}
 */
const npcDescs = {
  grandma: {
    id: "grandma",
  },
  farmhand: {
    id: "farmhand",
  },
  jack: {
    id: "jack",
  },
  blacksmith: {
    id: "blacksmith",
  },
  knightR: {
    id: "knightR",
  },
  knightG: {
    id: "knightG",
  },
  knightY: {
    id: "knightY",
  },
  knightB: {
    id: "knightB",
  },
  mom: {
    id: "mom",
  },
  bro: {
    id: "bro",
  },
  hermit: {
    id: "hermit",
  },
  sheep: {
    id: "sheep",
  },
  wizard: {
    id: "wizard",
  },
  bard: {
    id: "bard",
  },
  conan: {
    id: "conan",
  },
  broody: {
    id: "broody",
  },
  gudul: {
    id: "gudul",
  },
  pup: {
    id: "pup",
  },
  samurai: {
    id: "samurai",
  },
  captain: {
    id: "captain",
  },
  onion: {
    id: "onion",
    uncounted: true,
  },
  salad: {
    id: "salad",
    uncounted: true,
  },
  leek: {
    id: "leek",
    uncounted: true,
  },
}

/**
 * @enum {number}
 */
const NpcStates = {
  Idle: 0,
  ChaseBaxton: 1,
  Hit: 2,
  CatchUp: 3,
  ChaseBaddie: 4,
  Attack: 5
};

/**
 * @typedef NPC
 * @type {{
     id: string,
     desc: NPCDesc,
     pos: PIXI.Point,
     collideBox: { width: number; height: number };
     hitbox: { width: number; height: number };

     speed: number;
     health: number;
     state: NpcStates;
     stateTimer: number;
     stateTarget: Baddie;

     knockbackDirection: PIXI.Point;

     pixiRoot: PIXI.Container,
     sprite: PIXI.AnimatedSprite,
     spritesheet: Spritesheet,
     shadowSprite: PIXI.Sprite,
     outlineFilter: PIXI.filters.OutlineFilter;

     following: boolean;
 * }}
 */

/**
 * @param {number} dt
 */
function updateNPCs(dt) {
  const { baxton } = inGame;
  const { hovered } = inGame.stats;

  for (let npc of inGame.activeWorld.npcs) {
    let isHovered = (hovered && hovered.kind == "npc" && hovered.id == npc.id);
    npc.outlineFilter.thickness = (isHovered ? 1 : 0) * game.stageScale;


    if (!npc.following) continue;

    if (npc.state === NpcStates.Hit) {
      npc.stateTimer -= dt;
      if (npc.stateTimer <= 0) {
        npc.sprite.loop = true;
        setNpcIdle(npc);
      } else {
        moveWithCollision(npc, npc.knockbackDirection);
        npc.knockbackDirection.x *= 0.9;
        npc.knockbackDirection.y *= 0.9;
      }
    }

    if (npc.desc.id === "salad") {
      let foundBaddieNearby = false;

      for (let baddie of inGame.activeWorld.baddies) {
        const dist = distxy(baddie.pos, npc.pos);
        if (dist < 8) {
          foundBaddieNearby = true;
          break;
        }
      }

      if (foundBaddieNearby) {
        setNpcIdle(npc);
        continue;
      }
    } else if (npc.desc.id === "leek") {
      if (npc.state === NpcStates.Attack) {
        npc.stateTimer -= dt;
        if (npc.stateTimer <= 0) {
          if (inGame.activeWorld.baddies.indexOf(npc.stateTarget) !== -1) {
            const diff = { x: npc.stateTarget.pos.x - npc.pos.x, y: npc.stateTarget.pos.y - npc.pos.y };

            const attackDamage = 2;

            if (Math.abs(diff.x) < npc.collideBox.width + npc.stateTarget.hitbox.width &&
              Math.abs(diff.y) < npc.collideBox.height + npc.stateTarget.hitbox.height) {
              const attackAngle = Math.atan2(-diff.y, -diff.x);
              damageBaddie(npc.stateTarget, attackDamage, attackAngle);
            }
          }

          npc.stateTarget = null;

          npc.sprite.loop = true;
          setNpcIdle(npc);
          continue;
        }
      } else {
        if (npc.state === NpcStates.ChaseBaddie) {
          if (inGame.activeWorld.baddies.indexOf(npc.stateTarget) === -1) {
            npc.stateTarget = null;
            setNpcIdle(npc);
            continue;
          }

          npc.stateTimer += dt;

          const diff = { x: npc.stateTarget.pos.x - npc.pos.x, y: npc.stateTarget.pos.y - npc.pos.y };
          const dist = lengthxy(diff);

          const maxChaseDelay = 3;
          const stopChaseRange = 10;

          if (Math.abs(diff.x) < npc.collideBox.width + npc.stateTarget.hitbox.width * 0.9 &&
            Math.abs(diff.y) < npc.collideBox.height + npc.stateTarget.hitbox.height * 0.9) {

            setNpcAttack(npc);
          } else if (dist > stopChaseRange || npc.stateTimer > maxChaseDelay) {
            setNpcIdle(npc);
            continue;
          } else {
            const angle = Math.atan2(diff.y, diff.x);
            const amount = Math.min(dist - npc.collideBox.width + npc.stateTarget.collideBox.width, npc.speed * dt);
            const move = { x: Math.cos(angle) * amount, y: Math.sin(angle) * amount };
            setNpcDirection(npc, move.x < 0);
            moveWithCollision(npc, move);
          }
        } else if (npc.state === NpcStates.ChaseBaxton || (npc.state === NpcStates.Idle && npc.stateTimer <= 0)) {
          for (let baddie of inGame.activeWorld.baddies) {
            const chaseRange = 7;
            const dangerousRange = 9;

            if (distxy(npc.pos, baddie.pos) <= chaseRange && distxy(baxton.pos, baddie.pos) <= dangerousRange) {
              setNpcChaseBaddie(npc, baddie);
              break;
            }
          }
        }
      }
    }

    // Common follow behavior
    const diff = { x: baxton.pos.x - npc.pos.x, y: baxton.pos.y - npc.pos.y };
    const angle = Math.atan2(diff.y, diff.x);
    const dist = lengthxy(diff);

    const tooFarRange = 15;
    const sleepRange = 2;

    switch (npc.state) {
      case NpcStates.Idle: {
        npc.stateTimer -= dt;

        if (npc.stateTimer <= 0 && dist > sleepRange && dist < tooFarRange) {
          setNpcChaseBaxton(npc);
        }
        break;
      }
      case NpcStates.ChaseBaxton: {
        if (dist <= sleepRange) {
          setNpcIdle(npc);
        } else if (dist >= tooFarRange) {
          setNpcCatchUp(npc);
        } else {
          const amount = Math.min(dist - npc.collideBox.width + baxton.collideBox.width, npc.speed * dt);
          const move = { x: Math.cos(angle) * amount, y: Math.sin(angle) * amount };
          setNpcDirection(npc, move.x < 0);
          moveWithCollision(npc, move);
        }

        break;
      }
      case NpcStates.CatchUp: {
        if (dist <= 1) {
          setNpcIdle(npc);
        } else {
          const catchUpSpeed = 16;
          const amount = Math.min(dist - npc.collideBox.width + baxton.collideBox.width, catchUpSpeed * dt);
          npc.pos.x += Math.cos(angle) * amount;
          npc.pos.y += Math.sin(angle) * amount;
        }

        break;
      }
    }
  }

  for (const npc of inGame.activeWorld.npcs) {
    if (npc.state !== NpcStates.CatchUp) {
      for (const otherNpc of inGame.activeWorld.npcs) {
        if (npc !== otherNpc && npc.following && otherNpc.following && otherNpc.state !== NpcStates.CatchUp) {
          pushaway(npc, otherNpc, dt);
        }
      }
    }

    if (npc.following) {
      // baxton pushes npcs away
      pushaway(npc, inGame.baxton, dt, false);
    } else {
      // non-follower npcs push baxton away
      pushaway(inGame.baxton, npc, dt, false);
    }
  }

  for (const npc of inGame.activeWorld.npcs) {
    npc.pixiRoot.position.copyFrom(npc.pos);
    npc.pixiRoot.zIndex = npc.pos.y;

    npc.shadowSprite.position.copyFrom(npc.pos);
    npc.shadowSprite.zIndex = npc.pos.y - 10;

    npc.sprite.update(app.ticker.deltaTime);
  }
}

/**
 *
 * @param {NPC} npc
 * @param {boolean} isLookingLeft
 */
function setNpcDirection(npc, isLookingLeft) {
  npc.sprite.scale.x = isLookingLeft ? -1 : 1;
}

/**
 * @param {World} world
 * @param {PointLike} pos
 * @param {NPCDesc} desc
 */
function createNPC(world, pos, desc, following = false) {
  const pixiRoot = new PIXI.Container();
  pixiRoot.scale.set(inverseTileSize);
  pixiRoot.position.set(pos.x, pos.y);
  world.pixiLayers["ground"].entitiesContainer.addChild(pixiRoot);

  const spritesheet = getSpritesheet(`inGame/npcs/${desc.id}/body`);

  const sprite = new PIXI.AnimatedSprite(spritesheet.animations["idle"], false);
  pixiRoot.addChild(sprite);
  sprite.anchor.copyFrom(spritesheet.anchor);
  sprite.play();
  const outlineFilter = new PIXI.filters.OutlineFilter(0, 0xffffff, 0.5);
  sprite.filters = [outlineFilter];

  const shadowSprite = new PIXI.Sprite(tex("inGame/shadow"));
  shadowSprite.scale.set(inverseTileSize * 1.2);
  shadowSprite.anchor.set(0.5, 0);
  shadowSprite.position.set(pos.x, pos.y);
  world.pixiShadowsLayer.addChild(shadowSprite);

  /** @type {NPC} */
  const npc = {
    id: generateID(),
    desc,
    pos: new PIXI.Point(pos.x, pos.y),
    collideBox: { width: 0.8, height: 0.45 },
    hitbox: { width: 0.9, height: 1.4 },

    speed: 6,
    health: 2,
    state: NpcStates.Idle,
    stateTimer: 0,
    stateTarget: null,
    knockbackDirection: new PIXI.Point(),

    pixiRoot,
    sprite,
    spritesheet,
    shadowSprite,
    outlineFilter,

    following
  };

  world.npcs.push(npc);
}

/**
 * @param {NPC} npc
 */
function setNpcIdle(npc) {
  npc.state = NpcStates.Idle;
  if (npc.desc.uncounted) {
    // vegetable npcs get tired
    npc.stateTimer = 0.7;
  } else {
    npc.stateTimer = 0.4;
  }

  setAnimation(npc.sprite, npc.spritesheet.animations.idle);
}

/**
 * @param {NPC} npc
 */
function setNpcChaseBaxton(npc) {
  npc.state = NpcStates.ChaseBaxton;

  setAnimation(npc.sprite, npc.spritesheet.animations.run);
}

/**
 * @param {NPC} npc
 * @param {Baddie} target
 */
function setNpcChaseBaddie(npc, target) {
  npc.state = NpcStates.ChaseBaddie;
  npc.stateTimer = 0;
  npc.stateTarget = target;

  setAnimation(npc.sprite, npc.spritesheet.animations.run);
}

/**
 * @param {NPC} npc
 */
function setNpcAttack(npc) {
  npc.state = NpcStates.Attack;
  npc.stateTimer = 0.3;

  npc.sprite.loop = false;
  setAnimation(npc.sprite, npc.spritesheet.animations.attack);
}

/**
 * @param {NPC} npc
 * @param {number} angle
 */
function setNpcHit(npc, angle) {
  npc.state = NpcStates.Hit;
  npc.stateTimer = 0.3;

  npc.knockbackDirection.set(-Math.cos(angle) * 0.05, -Math.sin(angle) * 0.05);

  npc.sprite.loop = false;
  setAnimation(npc.sprite, npc.spritesheet.animations.hit);
}

/**
 * @param {NPC} npc 
 */
function setNpcCatchUp(npc) {
  npc.state = NpcStates.CatchUp;

  setAnimation(npc.sprite, npc.spritesheet.animations.run);
}

/**
 * @param {NPC} npc
 * @param {number} damage
 * @param {number} angle
 */
function damageNpc(npc, damage, angle) {
  npc.health = Math.max(npc.health - damage, 0);

  play2Dsound(`${npc.health === 0 ? "deaths" : "hits"}.npcs.${npc.desc.id}`, npc.pos);

  if (npc.health === 0) {
    destroyNpc(npc);
    spawnAnimatedFx("inGame/fxs/vanish", npc.pos, 0);
  } else {
    setNpcHit(npc, angle);
  }
}

/**
 * @param {NPC} npc
 */
function destroyNpc(npc) {
  inGame.activeWorld.pixiLayers["ground"].entitiesContainer.removeChild(npc.pixiRoot);
  inGame.activeWorld.pixiShadowsLayer.removeChild(npc.shadowSprite);
  inGame.activeWorld.npcs.splice(inGame.activeWorld.npcs.indexOf(npc), 1);

  npc.sprite.destroy();
  npc.shadowSprite.destroy();
}
